import os
from pathlib import Path
import pandas as pd
import duckdb
import json
import streamlit as st


def repr_single(s):
    return "'" + repr('"' + s)[2:]

@st.cache_resource
def read_files(folder):
    home = Path.home()
    lookup_dir = os.path.join(home, "Documents", "Cast", "data", folder)
    files = os.listdir(lookup_dir)
    main_view = dict()

    for file in files:
        if file.endswith(".DS_Store") is False:
            if file.endswith("json") is False:
                lookup_file = os.path.join(lookup_dir, file)
                sample = duckdb.sql(f"SELECT * FROM '{lookup_file}' limit 5000").df()
                main_view[file.split(".")[0]] = sample
                get_conversion_mapping(main_view, file.split(".")[0])

    return main_view


@st.cache_resource
def process_types(main_view, selected_object):
    home = Path.home()
    lookup_dir = os.path.join(home, "Documents", "Cast", "settings.json")
    with open(lookup_dir, "r") as f:
        settings = json.load(f)

    head = main_view[selected_object].head(1).T
    head.columns = ["Sample"]
    dtype = pd.DataFrame(main_view[selected_object].dtypes)
    dtype.columns = ["Type"]
    dtype["Type"] = dtype["Type"].apply(lambda x: "string" if x == "object" else str(x))

    df = main_view[selected_object]
    for col in df.select_dtypes(include="object").columns:
        df[col] = df[col].replace(settings["missing_values"], "NULL")

    df = df.fillna(0)
    df_numeric = df.apply(pd.to_numeric, errors="coerce")

    percent_nulls = pd.DataFrame(100 - df_numeric.isna().mean() * 100)
    percent_nulls.columns = ["Perc_numeric"]
    temp_1 = dtype.join(head)
    result = temp_1.join(percent_nulls)

    result["Sample"] = result["Sample"].astype("str")

    return result


@st.cache_resource
def get_conversion_mapping(main_view, selected_object):
    statements = []
    type_df = process_types(main_view, selected_object)

    numeric_casts = list(type_df[(type_df.Perc_numeric > 0.2) & (type_df.Type == "string")].index)
    home = Path.home()
    lookup_dir = os.path.join(home, "Documents", "Cast", "settings.json")
    with open(lookup_dir, "r") as f:
        settings = json.load(f)

    missing = settings["missing_values"]
    numeric_separator = settings["numeric_separator"]
    string_casts = list(type_df[(type_df.Perc_numeric < 0.2) & (type_df.Type == "string")].index)

    all_columns = list(type_df.index)

    for col in all_columns:
        if col not in string_casts:
            if col not in numeric_casts:
                statement = col
                statements.append(statement)

    for entry in numeric_casts:
        values = ','.join(repr_single(w) for w in missing)
        statement = f"""coalesce(
TRY_CAST(IF({entry}::VARCHAR in ({values}),'',if({entry}::VARCHAR ilike '%{numeric_separator}%',replace({entry}::VARCHAR,'{numeric_separator}','.'),{entry}::VARCHAR)) as Float) 
,TRY_CAST({entry} as Float)) as  {entry}"""
        statements.append(statement)

    for entry in string_casts:
        values = ','.join(repr_single(w) for w in missing)  # repr() ensures correct quoting
        statement = f"IF({entry} in ({values}), NULL, {entry}) as {entry}"
        statements.append(statement)

    home = Path.home()

    lookup_dir = os.path.join(home, "Documents", "Cast", "data", "source")

    data_source=''
    files = os.listdir(lookup_dir)
    for file in files:
        if file.startswith(selected_object):
            data_source = file
            break
        else:
            data_source = f"{selected_object}.parquet"


    if data_source=='':
        return None
    
    clean_statements = "\n \t,".join(statements)
    file = os.path.join(home, "Documents", "Cast", "data", "source", data_source)

    final_query = f"""
select
    {clean_statements}
from '{file}'
    """

    lookup_dir = os.path.join(home, "Documents", "Cast", "queries")
    if f"{selected_object}.sql" not in os.listdir(lookup_dir):
        file_dir = os.path.join(home, "Documents", "Cast", "queries", f"{selected_object}.sql")
        with open(file_dir, "w") as f:
            f.write(final_query)


@st.cache_resource
def get_numeric_cols(main_view, selected_object):
    df = process_types(main_view, selected_object)
    df["Type"] = df["Type"].astype("str")
    return list(df[df["Type"].isin(["float64", "float32,", "int64", "int32"])].index)



import streamlit as st
from streamlit_monaco import st_monaco
import duckdb
from pathlib import Path
import os
from utils.files import delete_file, list_files, load_file, save_file, remove_incomplete_rows


@st.dialog("Save query")
def save_modal(filename, text):
    st.write("Choose a filename:")
    new_name = st.text_input(value=filename, label="Filename:")
    if st.button("Save"):
        save_file(file_name=f"{new_name.split('.')[0]}.sql", text=text)
        st.rerun()


@st.dialog("Delete")
def delete_modal(filename):
    st.write(f"Are you sure you want to delete {filename}")
    if st.button("Yes"):
        delete_file(filename)
        st.rerun()
        
@st.dialog("Upload to Exakt")
def upload_modal():
    workspaces=['Belgium - MTPL']
    st.write('''I confirm that the data l'm uploading does not contain any Personal Identifiable Information (PI).
PlI includes but is not limited to names, addresses, social security numbers, phone numbers, email addresses, and any other information that could be used to identify an individual.
    ''')
    upload_check=st.checkbox("I understand and confirm my data does not contain PII")
    
    if upload_check:
        with st.form("dataset_form"):
            selected_workspace = st.selectbox("Select Workspace", options=workspaces)
            datasource_type = st.selectbox("Select Datasource Type", options=["Geo", "Modelling"])
            dataset_name = st.text_input("Dataset Name", placeholder="Enter a name for your dataset")
            submit_button = st.form_submit_button("Upload Dataset")
            
        if submit_button:
            st.write("Submitted")
            st.rerun()

@st.dialog("Export")
def export_modal(selected_file, alternative):
    st.write("Do you want to exclude empty entries?")

    include_empty = st.radio("Include empty entries", ["Yes", "No"])

    if st.button("Run"):
        home = Path.home()
        lookup_dir = os.path.join(home, "Documents", "Cast", "data", "export", selected_file.split(".")[0])
        with st.spinner(text="In progress..."):
            sql_rendered = f"SET ieee_floating_point_ops = false; \n {alternative}"
            data = duckdb.sql(sql_rendered).pl()

            if include_empty == "Yes":
                data.write_parquet(f"{lookup_dir}.parquet")
            else:
                remove_incomplete_rows(data).write_parquet(f"{lookup_dir}.parquet")
            st.rerun()


def query_extract():

    if "ran" not in st.session_state:
        st.session_state["ran"] = False
    c1, c2 = st.columns([1, 9], gap="small")
    upload_button = False
    with c1:
        if len(list_files()) > 0:
            selected_file = st.selectbox(options=list_files(), label="Query")
            if "selected_file" not in st.session_state:
                st.session_state["selected_file"] = selected_file
                text = load_file(selected_file)
                st.session_state["query"] = text

            if st.session_state["selected_file"] != selected_file:
                st.session_state["selected_file"] = selected_file
                text = load_file(selected_file)
                st.session_state["query"] = text

            else:
                text = st.session_state["query"]

            run_button = st.button("Run", type="primary", use_container_width=True)
            export_button = st.button("Export", type="secondary", use_container_width=True)
            save_button = st.button("Save", type="secondary", use_container_width=True)
            delete_button = st.button("Delete", type="secondary", use_container_width=True)
            
            if 'exakt' in st.session_state["settings"]:
                upload_button = st.button("Upload", icon="🚀", use_container_width=True)
            else:
                upload_button = False
        else:
            selected_file = "new_query.sql"
            text = ""
            run_button = st.button("Run", type="primary", use_container_width=True, disabled=True)
            export_button = st.button("Export", type="secondary", use_container_width=True, disabled=True)
            save_button = st.button("Save", type="secondary", use_container_width=True)
            delete_button = st.button("Delete", type="secondary", use_container_width=True, disabled=True)

    with c2:
        alternative = st_monaco(value=text, height="500px", language="sql", theme="vs-dark", lineNumbers=True)

    if run_button:
        try:
            t1, t2 = st.tabs(["Results", "Freestyle"])
            sql_rendered = f"select * from ({alternative}) limit 1000"
            st.session_state["ran"] = True

            with t1:
                if run_button:
                    with st.spinner("Refreshing"):
                        st.session_state["ran"]
                        df = duckdb.sql(sql_rendered).df()
                        st.session_state["df"] = df
                else:
                    df = st.session_state["df"]

                st.dataframe(df.head(1000))
        except Exception as e:
            st.write(e)

    elif save_button:
        try:
            save_modal(selected_file, alternative)
        except Exception as e:
            st.write(e)

    elif delete_button:
        delete_modal(filename=selected_file)

    elif export_button:
        export_modal(selected_file, alternative)
        
    elif upload_button:
        upload_modal(selected_file)



import os
from pathlib import Path
import pandas as pd
import duckdb
from pygwalker.api.streamlit import StreamlitRenderer
import streamlit as st


from utils.config import read_setting_local


@st.cache_resource
def get_pyg_renderer(file_path) -> StreamlitRenderer:
    settings = read_setting_local()
    try:
        data = duckdb.sql(f"SELECT * FROM '{file_path}.parquet' limit {settings['interactive_limit']}").df()
        data = data.astype({col: str for col in data.select_dtypes(include=["datetime"])})
    except:
        try:
            data = duckdb.sql(f"SELECT * FROM '{file_path}.csv' limit {settings['interactive_limit']}").df()
            data = data.astype({col: str for col in data.select_dtypes(include=["datetime"])})
            
        except:
            data=pd.DataFrame()
            
    home = Path.home()

    filepath = Path(file_path)
    last_entry = filepath.name
    second_last_entry = filepath.parent.name

    file_dir = os.path.join(home, "Documents", "Cast", "configs", f"{second_last_entry}|{last_entry}.json")

    if len(data)>0:
        return StreamlitRenderer(data, spec=file_dir, spec_io_mode="rw", kernel_computation=True)


def walker_viewer(main_view, export_view):
    if "ran" not in st.session_state:
        st.session_state["ran"] = False

    c1, c2 = st.columns([3, 17], gap="small")

    with c1:
        folder_set = st.selectbox("Folder", ["Source", "Export"], key="dataset_dw_walker")
        if folder_set == "Source":
            if len(main_view.keys()) > 0:
                data_set = st.selectbox(options=main_view.keys(), label="DataSource", key="export_ds_walker")
                submitted = st.button("Explore")
            else:
                data_set=''
                submitted = False
        else:
            if len(export_view.keys()) > 0:
                data_set = st.selectbox(options=export_view.keys(), label="DataSource", key="export_ds_walker")
                submitted = st.button("Explore")
            else:
                data_set=''
                submitted = False

        if "py_walker" not in st.session_state:
            if submitted:
                
                st.session_state["py_walker"] = f"{folder_set}|{data_set}"
            else:
                st.session_state["py_walker"] = ""

        if not submitted:
            if st.session_state["py_walker"] != f"{folder_set}|{data_set}":
                st.session_state["py_show"] = False
            else:
                st.session_state["py_show"] = True
        else:
            st.session_state["py_walker"] = f"{folder_set}|{data_set}"
            st.session_state["py_show"] = True

    with c2:
        home = Path.home()
        base_dir = os.path.join(home, "Documents", "Cast", "data")

        try:
            pyg_app = get_pyg_renderer(file_path=f"{base_dir}/{folder_set.lower()}/{data_set}")
            pyg_app.explorer()
        except:
            st.write("No data to display")


import streamlit as st
from components.bi_tool import walker_viewer
from components.extract import query_extract
from utils.config import read_setting
from utils.data_profiler import get_numeric_cols, read_files
from utils.documentation import parse_documentation
from components.graph import make_graph


st.set_page_config(
    page_title="Exakt Cast",
    page_icon=None,
    layout="wide",
    initial_sidebar_state="auto",
    menu_items=None,
)
hide_streamlit_style = """
<style>
#MainMenu {visibility: hidden;}
footer {visibility: hidden;}
stAppDeployButton {visibility: hidden;}
header {visibility: hidden;}
</style>
"""
st.markdown(hide_streamlit_style, unsafe_allow_html=True)


try:
    st.image(
        image="assets/cast.png",
        caption=None,
        width=200,
        use_column_width=None,
        clamp=False,
        channels="RGB",
        output_format="auto",
    )
except:
    st.image(
        image="cast/assets/cast.png",
        caption=None,
        width=200,
        use_column_width=None,
        clamp=False,
        channels="RGB",
        output_format="auto",
    )

with st.spinner("Refreshing"):
    main_view = read_files(folder="source")
    export_view = read_files(folder="export")
    documentation = parse_documentation()
    settings = read_setting()
    
    if settings not in st.session_state:
        st.session_state["settings"] = settings

main_col1, main_col2 = st.columns([8, 2])


with main_col2:
    clear_cache_btn = st.button("Refresh", icon="🔄", type="tertiary")
    if clear_cache_btn:
        st.cache_data.clear()
        st.cache_resource.clear()

q_t1, q_t2, q_t3, q_t4 = st.tabs(["Data", "One-way", "Data Viewer", "Help"])

with q_t1:
    query_extract()

with q_t2:
    q_t1c1, q_t1c2 = st.columns([3, 17], gap="small")

    with q_t1c1:
        folder_set = st.selectbox("Folder", ["Source", "Export"])

        if folder_set == "Source":
            if len(main_view.keys()) > 0:
                data_set = st.selectbox(options=main_view.keys(), label="Datasource")
                data_dim = st.selectbox(options=list(main_view[data_set].columns), label="Dimension")
                data_target = st.selectbox(options=get_numeric_cols(main_view, data_set), label="Target")
                submitted = st.button("render")

            else:
                submitted = False
        else:
            if len(export_view.keys()) > 0:
                data_set = st.selectbox(options=export_view.keys(), label="Datasource")
                data_dim = st.selectbox(options=list(export_view[data_set].columns), label="Dimension")
                data_target = st.selectbox(options=get_numeric_cols(export_view, data_set), label="Target")
                submitted = st.button("render")
            else:
                submitted = False

    if submitted:
        with q_t1c2:
            st.plotly_chart(make_graph(folder_set, data_set, data_dim, data_target))


with q_t3:
    walker_viewer(main_view, export_view)

with q_t4:
    documentation_list = st.selectbox(options=documentation.keys(), label="Topic")
    st.divider()
    st.markdown(documentation[documentation_list])



import plotly.graph_objects as go
import duckdb as db
import os
from pathlib import Path


def make_graph(folder_set, data_set, data_dim, data_target):
    home = Path.home()
    lookup_dir = os.path.join(home, "Documents", "Cast", "data", folder_set.lower(), data_set)

    try:
        query = f"""
        select  coalesce({data_dim},'-1') as {data_dim},avg({data_target}) as {data_target},count(*) as volumes
        from  '{lookup_dir}.csv' 
        group by 1
        order by 1
        
        """
        data_extract = db.sql(query=query).df()

    except:
        query = f"""
        select  coalesce({data_dim},'-1') as {data_dim},avg({data_target}) as {data_target},count(*) as volumes
        from  '{lookup_dir}.parquet' 
        group by 1
        order by 1
        """

        data_extract = db.sql(query=query).df()

    # Create a bar chart for volumes
    bar_chart = go.Bar(x=data_extract[data_dim], y=data_extract["volumes"], name="Volume", marker_color="rgba(55, 83, 109, 0.7)")

    # Create a line chart for average amounts
    line_chart = go.Scatter(
        x=data_extract[data_dim],
        y=data_extract[f"{data_target}"],
        name=f"Average {data_target}",
        mode="lines+markers",
        marker=dict(color="#3dbf79", size=8),
        line=dict(color="#3dbf79", width=3),
        yaxis="y2",
    )

    # Combine both charts in a single figure
    fig = go.Figure(data=[bar_chart, line_chart])

    # Customize the layout
    fig.update_layout(
        title=f"Volumes and Average {data_target}",
        xaxis_title=f"{data_dim}",
        yaxis_title="Values",
        yaxis=dict(title="Volume", side="left"),
        yaxis2=dict(title=f"Average  {data_target}", overlaying="y", side="right"),
        barmode="group",
        hovermode="x unified",
        legend=dict(x=0.1, y=1.1),
    )

    return fig



import os
import streamlit as st


@st.cache_data
def parse_documentation():
    os.getcwd()  # noqa: F821
    try:
        files = os.listdir(f"{os.getcwd()}/assets/")
        documentation = dict()
        for file in files:
            if file.endswith("md"):
                content = open(f"{os.getcwd()}/assets/{file}").read()
                documentation[file.split(".")[0]] = content

    except Exception as e:
        files = os.listdir(f"{os.getcwd()}/assets/")
        documentation = dict()

        for file in files:
            if file.endswith("md"):
                content = open(f"{os.getcwd()}/assets/{file}").read()
                documentation[file.split(".")[0]] = content

    return documentation



from pydantic import BaseModel
import boto3

class UploadS3(BaseModel):
    endpoint: str
    bucket: str
    access_key: str
    secret_key: str
    source_file: str
    dest_file: str


def upload_data(config: UploadS3):
    s3_endpoint = config.endpoint
    s3_bucket = config.bucket
    s3_access_key = config.access_key
    s3_secret_key = config.secret_key

    s3_client = boto3.client(
        service_name="s3",
        endpoint_url=s3_endpoint,
        aws_access_key_id=s3_access_key,
        aws_secret_access_key=s3_secret_key,
        region_name="auto"
    )

    response = s3_client.upload_file(config.source_file, s3_bucket,config.dest_file)
    return response


from pathlib import Path
import subprocess
import os
import json


def main():
    os.chdir(os.path.dirname(__file__))
    home = Path.home()
    base_dir = os.path.join(home, "Documents", "Cast")
    sub_dirs = [
        os.path.join(base_dir, "data", "source"),
        os.path.join(base_dir, "data", "export"),
        os.path.join(base_dir, "queries"),
         os.path.join(base_dir, "configs"),
    ]
    for sub_dir in sub_dirs:
        os.umask(0)
        os.makedirs(os.path.join(base_dir, sub_dir), exist_ok=True,mode=0o777)

    json_content = {
        "missing_values": ["NaN", "None", "", " ", "\t", "N/A", "NULL", "NA", "n/a", "missing", "unknown", "Undefined", "-", "--", "?"],
        "numeric_separator": ",",
        "interactive_limit": 2000000,
    }

    # Path to the JSON file
    json_file_path = os.path.join(base_dir, "settings.json")

    if not os.path.exists(json_file_path):
        # Write the content to the JSON file
        with open(json_file_path, "w") as json_file:
            json.dump(json_content, json_file, indent=4)

    subprocess.run(["streamlit", "run", "app.py"])



import os
from pathlib import Path
import polars as pl


def list_files():
    home = Path.home()
    lookup_dir = os.path.join(home, "Documents", "Cast", "queries")
    files = os.listdir(lookup_dir)

    new_files = []
    for file in files:
        if file.endswith(".sql"):
            new_files.append(file)

    return new_files


def first_file():
    home = Path.home()
    lookup_dir = os.path.join(home, "Documents", "Cast", "queries")
    files = os.listdir(lookup_dir)
    for file in files:
        if file.endswith(".DS_Store"):
            continue
        else:
            return file


def load_file(file_name):
    home = Path.home()
    file_dir = os.path.join(home, "Documents", "Cast", "queries", file_name)
    with open(file_dir, "r") as f:
        text = f.read()
    return text


def save_file(file_name, text):
    home = Path.home()
    file_dir = os.path.join(home, "Documents", "Cast", "queries", file_name)
    with open(file_dir, "w") as f:
        f.write(text)


def delete_file(file):
    home = Path.home()
    file_dir = os.path.join(home, "Documents", "Cast", "queries", file)
    os.remove(file_dir)


def remove_incomplete_rows(df: pl.DataFrame, threshold: float = 0.54) -> pl.DataFrame:
    """
    Removes rows from a Polars DataFrame with more than the specified percentage of null values.

    Parameters:
        df (pl.DataFrame): The input Polars DataFrame.
        threshold (float): The fraction of null values allowed per row (default is 0.5).

    Returns:
        pl.DataFrame: A DataFrame with rows removed based on the null value threshold.
    """
    # Calculate the maximum allowed nulls per row
    null_limit = int(threshold * len(df.columns))

    schema = df.schema

    # Iterate through columns and apply NaN replacement for numeric types only
    filtered_df = (
        df.with_columns(
            [
                # Apply the transformation only for numeric columns
                pl.when(pl.col(col).is_nan()).then(None).otherwise(pl.col(col)).alias(col)
                if schema[col] in [pl.Float32, pl.Float64]
                else pl.col(col)
                for col in df.columns
            ]
        )
        .with_columns(sum([pl.col(col).is_null().cast(int) for col in df.columns]).alias("null_count"))
        .filter(pl.col("null_count") <= null_limit)
        .drop("null_count")
    )  # Remove the temporary null count column

    return filtered_df



import json
from pathlib import Path
import streamlit as st
import os


@st.cache_data
def read_setting():
    home = Path.home()
    file_dir = os.path.join(home, "Documents", "Cast", "settings.json")
    with open(file_dir, "r") as f:
        settings = json.load(f)
    return settings


def read_setting_local():
    home = Path.home()
    file_dir = os.path.join(home, "Documents", "Cast", "settings.json")
    with open(file_dir, "r") as f:
        settings = json.load(f)
    return settings
